@@ -10,7 +10,6 @@
 #include <typeinfo>
 #include <algorithm>
 #include <unordered_map>
-#include "sass/base.h"
 
 #ifdef __clang__
 
@@ -44,29 +43,15 @@
 #include "error_handling.hpp"
 #include "ast_def_macros.hpp"
 #include "ast_fwd_decl.hpp"
+#include "to_string.hpp"
 #include "source_map.hpp"
 
 #include "sass.h"
 
 namespace Sass {
 
-  // ToDo: should this really be hardcoded
-  // Note: most methods follow precision option
   const double NUMBER_EPSILON = 0.00000000000001;
 
-  // ToDo: where does this fit best?
-  // We don't share this with C-API?
-  class Operand {
-    public:
-      Operand(Sass_OP operand, bool ws_before = false, bool ws_after = false)
-      : operand(operand), ws_before(ws_before), ws_after(ws_after)
-      { }
-    public:
-      enum Sass_OP operand;
-      bool ws_before;
-      bool ws_after;
-  };
-
   // from boost (functional/hash):
   // http://www.boost.org/doc/libs/1_35_0/doc/html/hash/combine.html
   // Boost Software License - Version 1.0
@@ -81,18 +66,13 @@ namespace Sass {
   //////////////////////////////////////////////////////////
   // Abstract base class for all abstract syntax tree nodes.
   //////////////////////////////////////////////////////////
-  class AST_Node : public Memory_Object {
+  class AST_Node {
     ADD_PROPERTY(ParserState, pstate)
   public:
     AST_Node(ParserState pstate)
     : pstate_(pstate)
     { }
     virtual ~AST_Node() = 0;
-    virtual size_t hash() { return 0; }
-    virtual std::string inspect() const { return to_string({ INSPECT, 5 }); }
-    virtual std::string to_sass() const { return to_string({ TO_SASS, 5 }); }
-    virtual std::string to_string(Sass_Inspect_Options opt) const;
-    virtual std::string to_string() const;
     // virtual Block* block() { return 0; }
   public:
     void update_pstate(const ParserState& pstate);
@@ -148,26 +128,9 @@ namespace Sass {
     virtual bool is_false() { return false; }
     virtual bool operator== (const Expression& rhs) const { return false; }
     virtual void set_delayed(bool delayed) { is_delayed(delayed); }
-    virtual bool has_interpolant() const { return is_interpolant(); }
-    virtual bool is_left_interpolant() const { return is_interpolant(); }
-    virtual bool is_right_interpolant() const { return is_interpolant(); }
-    virtual std::string inspect() const { return to_string({ INSPECT, 5 }); }
-    virtual std::string to_sass() const { return to_string({ TO_SASS, 5 }); }
     virtual size_t hash() { return 0; }
   };
 
-  //////////////////////////////////////////////////////////////////////
-  // Still just an expression, but with a to_string method
-  //////////////////////////////////////////////////////////////////////
-  class PreValue : public Expression {
-  public:
-    PreValue(ParserState pstate,
-               bool d = false, bool e = false, bool i = false, Concrete_Type ct = NONE)
-    : Expression(pstate, d, e, i, ct)
-    { }
-    virtual ~PreValue() { }
-  };
-
   //////////////////////////////////////////////////////////////////////
   // base class for values that support operations
   //////////////////////////////////////////////////////////////////////
@@ -178,6 +141,7 @@ namespace Sass {
     : Expression(pstate, d, e, i, ct)
     { }
     virtual bool operator== (const Expression& rhs) const = 0;
+    virtual std::string to_string(bool compressed = false, int precision = 5) const = 0;
   };
 }
 
@@ -224,10 +188,9 @@ namespace Sass {
     virtual ~Vectorized() = 0;
     size_t length() const   { return elements_.size(); }
     bool empty() const      { return elements_.empty(); }
-    T last() const          { return elements_.back(); }
-    T first() const         { return elements_.front(); }
+    T last()                { return elements_.back(); }
+    T first()               { return elements_.front(); }
     T& operator[](size_t i) { return elements_[i]; }
-    virtual const T& at(size_t i) const { return elements_.at(i); }
     const T& operator[](size_t i) const { return elements_[i]; }
     Vectorized& operator<<(T element)
     {
@@ -251,16 +214,6 @@ namespace Sass {
     const std::vector<T>& elements() const { return elements_; }
     std::vector<T>& elements(std::vector<T>& e) { elements_ = e; return elements_; }
 
-    virtual size_t hash()
-    {
-      if (hash_ == 0) {
-        for (T& el : elements_) {
-          hash_combine(hash_, el->hash());
-        }
-      }
-      return hash_;
-    }
-
     typename std::vector<T>::iterator end() { return elements_.end(); }
     typename std::vector<T>::iterator begin() { return elements_.begin(); }
     typename std::vector<T>::const_iterator end() const { return elements_.end(); }
@@ -583,32 +536,26 @@ namespace Sass {
   // necessary to store a list of each in an Import node.
   ////////////////////////////////////////////////////////////////////////////
   class Import : public Statement {
-    std::vector<Expression*> urls_;
-    std::vector<Include>     incs_;
-    ADD_PROPERTY(List*,      media_queries);
+    std::vector<std::string>         files_;
+    std::vector<Expression*>    urls_;
+    ADD_PROPERTY(List*, media_queries);
   public:
     Import(ParserState pstate)
     : Statement(pstate),
+      files_(std::vector<std::string>()),
       urls_(std::vector<Expression*>()),
-      incs_(std::vector<Include>()),
       media_queries_(0)
     { statement_type(IMPORT); }
-    std::vector<Expression*>& urls() { return urls_; }
-    std::vector<Include>& incs() { return incs_; }
+    std::vector<std::string>&      files()    { return files_; }
+    std::vector<Expression*>& urls()     { return urls_; }
     ATTACH_OPERATIONS()
   };
 
-  // not yet resolved single import
-  // so far we only know requested name
   class Import_Stub : public Statement {
-    Include resource_;
+    ADD_PROPERTY(std::string, file_name)
   public:
-    std::string abs_path() { return resource_.abs_path; };
-    std::string imp_path() { return resource_.imp_path; };
-    Include resource() { return resource_; };
-
-    Import_Stub(ParserState pstate, Include res)
-    : Statement(pstate), resource_(res)
+    Import_Stub(ParserState pstate, std::string f)
+    : Statement(pstate), file_name_(f)
     { statement_type(IMPORT_STUB); }
     ATTACH_OPERATIONS()
   };
@@ -674,10 +621,6 @@ namespace Sass {
     If(ParserState pstate, Expression* pred, Block* con, Block* alt = 0)
     : Has_Block(pstate, con), predicate_(pred), alternative_(alt)
     { statement_type(IF); }
-    virtual bool has_content()
-    {
-      return Has_Block::has_content() || (alternative_ && alternative_->has_content());
-    }
     ATTACH_OPERATIONS()
   };
 
@@ -865,8 +808,8 @@ namespace Sass {
     std::string type() { return is_arglist_ ? "arglist" : "list"; }
     static std::string type_name() { return "list"; }
     const char* sep_string(bool compressed = false) const {
-      return separator() == SASS_SPACE ?
-        " " : (compressed ? "," : ", ");
+      return separator() == SASS_COMMA ?
+        (compressed ? "," : ", ") : " ";
     }
     bool is_invisible() const { return empty(); }
     Expression* value_at_index(size_t i);
@@ -891,6 +834,7 @@ namespace Sass {
     }
 
     virtual bool operator== (const Expression& rhs) const;
+    virtual std::string to_string(bool compressed = false, int precision = 5) const;
 
     ATTACH_OPERATIONS()
   };
@@ -923,49 +867,29 @@ namespace Sass {
     }
 
     virtual bool operator== (const Expression& rhs) const;
+    virtual std::string to_string(bool compressed = false, int precision = 5) const;
 
     ATTACH_OPERATIONS()
   };
 
-  inline static const std::string sass_op_to_name(enum Sass_OP op) {
-    switch (op) {
-      case AND: return "and"; break;
-      case OR: return "or"; break;
-      case EQ: return "eq"; break;
-      case NEQ: return "neq"; break;
-      case GT: return "gt"; break;
-      case GTE: return "gte"; break;
-      case LT: return "lt"; break;
-      case LTE: return "lte"; break;
-      case ADD: return "plus"; break;
-      case SUB: return "sub"; break;
-      case MUL: return "times"; break;
-      case DIV: return "div"; break;
-      case MOD: return "mod"; break;
-      // this is only used internally!
-      case NUM_OPS: return "[OPS]"; break;
-      default: return "invalid"; break;
-    }
-  }
-
   //////////////////////////////////////////////////////////////////////////
   // Binary expressions. Represents logical, relational, and arithmetic
   // operations. Templatized to avoid large switch statements and repetitive
   // subclassing.
   //////////////////////////////////////////////////////////////////////////
-  class Binary_Expression : public PreValue {
+  class Binary_Expression : public Expression {
   private:
-    ADD_HASHED(Operand, op)
-    ADD_HASHED(Expression*, left)
-    ADD_HASHED(Expression*, right)
+    ADD_PROPERTY(enum Sass_OP, type)
+    ADD_PROPERTY(Expression*, left)
+    ADD_PROPERTY(Expression*, right)
     size_t hash_;
   public:
     Binary_Expression(ParserState pstate,
-                      Operand op, Expression* lhs, Expression* rhs)
-    : PreValue(pstate), op_(op), left_(lhs), right_(rhs), hash_(0)
+                      enum Sass_OP t, Expression* lhs, Expression* rhs)
+    : Expression(pstate), type_(t), left_(lhs), right_(rhs), hash_(0)
     { }
     const std::string type_name() {
-      switch (type()) {
+      switch (type_) {
         case AND: return "and"; break;
         case OR: return "or"; break;
         case EQ: return "eq"; break;
@@ -979,38 +903,10 @@ namespace Sass {
         case MUL: return "mul"; break;
         case DIV: return "div"; break;
         case MOD: return "mod"; break;
-        // this is only used internally!
-        case NUM_OPS: return "[OPS]"; break;
+        case NUM_OPS: return "num_ops"; break;
         default: return "invalid"; break;
       }
     }
-    const std::string separator() {
-      switch (type()) {
-        case AND: return "&&"; break;
-        case OR: return "||"; break;
-        case EQ: return "=="; break;
-        case NEQ: return "!="; break;
-        case GT: return ">"; break;
-        case GTE: return ">="; break;
-        case LT: return "<"; break;
-        case LTE: return "<="; break;
-        case ADD: return "+"; break;
-        case SUB: return "-"; break;
-        case MUL: return "*"; break;
-        case DIV: return "/"; break;
-        case MOD: return "%"; break;
-        // this is only used internally!
-        case NUM_OPS: return "[OPS]"; break;
-        default: return "invalid"; break;
-      }
-    }
-    bool is_left_interpolant(void) const;
-    bool is_right_interpolant(void) const;
-    bool has_interpolant() const
-    {
-      return is_left_interpolant() ||
-             is_right_interpolant();
-    }
     virtual void set_delayed(bool delayed)
     {
       right()->set_delayed(delayed);
@@ -1036,13 +932,12 @@ namespace Sass {
     virtual size_t hash()
     {
       if (hash_ == 0) {
-        hash_ = std::hash<size_t>()(type());
+        hash_ = std::hash<size_t>()(type_);
         hash_combine(hash_, left()->hash());
         hash_combine(hash_, right()->hash());
       }
       return hash_;
     }
-    enum Sass_OP type() const { return op_.operand; }
     ATTACH_OPERATIONS()
   };
 
@@ -1053,8 +948,8 @@ namespace Sass {
   public:
     enum Type { PLUS, MINUS, NOT };
   private:
-    ADD_HASHED(Type, type)
-    ADD_HASHED(Expression*, operand)
+    ADD_PROPERTY(Type, type)
+    ADD_PROPERTY(Expression*, operand)
     size_t hash_;
   public:
     Unary_Expression(ParserState pstate, Type t, Expression* o)
@@ -1098,8 +993,8 @@ namespace Sass {
   // Individual argument objects for mixin and function calls.
   ////////////////////////////////////////////////////////////
   class Argument : public Expression {
-    ADD_HASHED(Expression*, value)
-    ADD_HASHED(std::string, name)
+    ADD_PROPERTY(Expression*, value)
+    ADD_PROPERTY(std::string, name)
     ADD_PROPERTY(bool, is_rest_argument)
     ADD_PROPERTY(bool, is_keyword_argument)
     size_t hash_;
@@ -1158,27 +1053,23 @@ namespace Sass {
       has_rest_argument_(false),
       has_keyword_argument_(false)
     { }
-
-    Argument* get_rest_argument();
-    Argument* get_keyword_argument();
-
     ATTACH_OPERATIONS()
   };
 
   //////////////////
   // Function calls.
   //////////////////
-  class Function_Call : public PreValue {
-    ADD_HASHED(std::string, name)
-    ADD_HASHED(Arguments*, arguments)
+  class Function_Call : public Expression {
+    ADD_PROPERTY(std::string, name)
+    ADD_PROPERTY(Arguments*, arguments)
     ADD_PROPERTY(void*, cookie)
     size_t hash_;
   public:
     Function_Call(ParserState pstate, std::string n, Arguments* args, void* cookie)
-    : PreValue(pstate), name_(n), arguments_(args), cookie_(cookie), hash_(0)
+    : Expression(pstate), name_(n), arguments_(args), cookie_(cookie), hash_(0)
     { concrete_type(STRING); }
     Function_Call(ParserState pstate, std::string n, Arguments* args)
-    : PreValue(pstate), name_(n), arguments_(args), cookie_(0), hash_(0)
+    : Expression(pstate), name_(n), arguments_(args), cookie_(0), hash_(0)
     { concrete_type(STRING); }
 
     virtual bool operator==(const Expression& rhs) const
@@ -1228,11 +1119,11 @@ namespace Sass {
   ///////////////////////
   // Variable references.
   ///////////////////////
-  class Variable : public PreValue {
+  class Variable : public Expression {
     ADD_PROPERTY(std::string, name)
   public:
     Variable(ParserState pstate, std::string n)
-    : PreValue(pstate), name_(n)
+    : Expression(pstate), name_(n)
     { }
 
     virtual bool operator==(const Expression& rhs) const
@@ -1265,8 +1156,8 @@ namespace Sass {
   public:
     enum Type { NUMBER, PERCENTAGE, DIMENSION, HEX };
   private:
-    ADD_HASHED(Type, type)
-    ADD_HASHED(std::string, value)
+    ADD_PROPERTY(Type, type)
+    ADD_PROPERTY(std::string, value)
     size_t hash_;
   public:
     Textual(ParserState pstate, Type t, std::string val)
@@ -1304,15 +1195,14 @@ namespace Sass {
   // Numbers, percentages, dimensions, and colors.
   ////////////////////////////////////////////////
   class Number : public Value {
-    ADD_HASHED(double, value)
+    ADD_PROPERTY(double, value)
     ADD_PROPERTY(bool, zero)
     std::vector<std::string> numerator_units_;
     std::vector<std::string> denominator_units_;
     size_t hash_;
   public:
     Number(ParserState pstate, double val, std::string u = "", bool zero = true);
-    bool zero() { return zero_; }
-    bool is_valid_css_unit() const;
+    bool            zero()              { return zero_; }
     std::vector<std::string>& numerator_units()   { return numerator_units_; }
     std::vector<std::string>& denominator_units() { return denominator_units_; }
     const std::vector<std::string>& numerator_units() const   { return numerator_units_; }
@@ -1321,9 +1211,8 @@ namespace Sass {
     static std::string type_name() { return "number"; }
     std::string unit() const;
 
-    bool is_unitless() const;
-    double convert_factor(const Number&) const;
-    bool convert(const std::string& unit = "", bool strict = false);
+    bool is_unitless();
+    void convert(const std::string& unit = "", bool strict = false);
     void normalize(const std::string& unit = "", bool strict = false);
     // useful for making one number compatible with another
     std::string find_convertible_unit() const;
@@ -1332,16 +1221,13 @@ namespace Sass {
     {
       if (hash_ == 0) {
         hash_ = std::hash<double>()(value_);
-        for (const auto numerator : numerator_units())
-          hash_combine(hash_, std::hash<std::string>()(numerator));
-        for (const auto denominator : denominator_units())
-          hash_combine(hash_, std::hash<std::string>()(denominator));
       }
       return hash_;
     }
 
     virtual bool operator< (const Number& rhs) const;
     virtual bool operator== (const Expression& rhs) const;
+    virtual std::string to_string(bool compressed = false, int precision = 5) const;
 
     ATTACH_OPERATIONS()
   };
@@ -1350,15 +1236,16 @@ namespace Sass {
   // Colors.
   //////////
   class Color : public Value {
-    ADD_HASHED(double, r)
-    ADD_HASHED(double, g)
-    ADD_HASHED(double, b)
-    ADD_HASHED(double, a)
+    ADD_PROPERTY(double, r)
+    ADD_PROPERTY(double, g)
+    ADD_PROPERTY(double, b)
+    ADD_PROPERTY(double, a)
+    ADD_PROPERTY(bool, sixtuplet)
     ADD_PROPERTY(std::string, disp)
     size_t hash_;
   public:
-    Color(ParserState pstate, double r, double g, double b, double a = 1, const std::string disp = "")
-    : Value(pstate), r_(r), g_(g), b_(b), a_(a), disp_(disp),
+    Color(ParserState pstate, double r, double g, double b, double a = 1, bool sixtuplet = true, const std::string disp = "")
+    : Value(pstate), r_(r), g_(g), b_(b), a_(a), sixtuplet_(sixtuplet), disp_(disp),
       hash_(0)
     { concrete_type(COLOR); }
     std::string type() { return "color"; }
@@ -1376,6 +1263,7 @@ namespace Sass {
     }
 
     virtual bool operator== (const Expression& rhs) const;
+    virtual std::string to_string(bool compressed = false, int precision = 5) const;
 
     ATTACH_OPERATIONS()
   };
@@ -1390,6 +1278,7 @@ namespace Sass {
     : Value(pstate), message_(msg)
     { concrete_type(C_ERROR); }
     virtual bool operator== (const Expression& rhs) const;
+    virtual std::string to_string(bool compressed = false, int precision = 5) const;
     ATTACH_OPERATIONS()
   };
 
@@ -1403,6 +1292,7 @@ namespace Sass {
     : Value(pstate), message_(msg)
     { concrete_type(C_WARNING); }
     virtual bool operator== (const Expression& rhs) const;
+    virtual std::string to_string(bool compressed = false, int precision = 5) const;
     ATTACH_OPERATIONS()
   };
 
@@ -1410,7 +1300,7 @@ namespace Sass {
   // Booleans.
   ////////////
   class Boolean : public Value {
-    ADD_HASHED(bool, value)
+    ADD_PROPERTY(bool, value)
     size_t hash_;
   public:
     Boolean(ParserState pstate, bool val)
@@ -1431,6 +1321,7 @@ namespace Sass {
     }
 
     virtual bool operator== (const Expression& rhs) const;
+    virtual std::string to_string(bool compressed = false, int precision = 5) const;
 
     ATTACH_OPERATIONS()
   };
@@ -1448,6 +1339,7 @@ namespace Sass {
     static std::string type_name() { return "string"; }
     virtual ~String() = 0;
     virtual bool operator==(const Expression& rhs) const = 0;
+    virtual std::string to_string(bool compressed = false, int precision = 5) const = 0;
     ATTACH_OPERATIONS()
   };
   inline String::~String() { };
@@ -1457,25 +1349,15 @@ namespace Sass {
   // evaluation phase.
   ///////////////////////////////////////////////////////////////////////
   class String_Schema : public String, public Vectorized<Expression*> {
-    // ADD_PROPERTY(bool, has_interpolants)
+    ADD_PROPERTY(bool, has_interpolants)
     size_t hash_;
   public:
     String_Schema(ParserState pstate, size_t size = 0, bool has_interpolants = false)
-    : String(pstate), Vectorized<Expression*>(size), hash_(0)
+    : String(pstate), Vectorized<Expression*>(size), has_interpolants_(has_interpolants), hash_(0)
     { concrete_type(STRING); }
     std::string type() { return "string"; }
     static std::string type_name() { return "string"; }
 
-    bool is_left_interpolant(void) const;
-    bool is_right_interpolant(void) const;
-    // void has_interpolants(bool tc) { }
-    bool has_interpolants() {
-      for (auto el : elements()) {
-        if (el->is_interpolant()) return true;
-      }
-      return false;
-    }
-
     virtual size_t hash()
     {
       if (hash_ == 0) {
@@ -1486,6 +1368,7 @@ namespace Sass {
     }
 
     virtual bool operator==(const Expression& rhs) const;
+    virtual std::string to_string(bool compressed = false, int precision = 5) const;
 
     ATTACH_OPERATIONS()
   };
@@ -1496,7 +1379,7 @@ namespace Sass {
   class String_Constant : public String {
     ADD_PROPERTY(char, quote_mark)
     ADD_PROPERTY(bool, can_compress_whitespace)
-    ADD_HASHED(std::string, value)
+    ADD_PROPERTY(std::string, value)
   protected:
     size_t hash_;
   public:
@@ -1524,7 +1407,7 @@ namespace Sass {
     }
 
     virtual bool operator==(const Expression& rhs) const;
-    virtual std::string inspect() const; // quotes are forced on inspection
+    virtual std::string to_string(bool compressed = false, int precision = 5) const;
 
     // static char auto_quote() { return '*'; }
     static char double_quote() { return '"'; }
@@ -1545,7 +1428,7 @@ namespace Sass {
       if (q && quote_mark_) quote_mark_ = q;
     }
     virtual bool operator==(const Expression& rhs) const;
-    virtual std::string inspect() const; // quotes are forced on inspection
+    virtual std::string to_string(bool compressed = false, int precision = 5) const;
     ATTACH_OPERATIONS()
   };
 
@@ -1682,7 +1565,8 @@ namespace Sass {
     { }
     bool exclude(std::string str)
     {
-      bool with = feature() && unquote(feature()->to_string()).compare("with") == 0;
+      To_String to_string;
+      bool with = feature() && unquote(feature()->perform(&to_string)).compare("with") == 0;
       List* l = static_cast<List*>(value());
       std::string v;
 
@@ -1691,7 +1575,7 @@ namespace Sass {
         if (!l || l->length() == 0) return str.compare("rule") != 0;
         for (size_t i = 0, L = l->length(); i < L; ++i)
         {
-          v = unquote((*l)[i]->to_string());
+          v = unquote((*l)[i]->perform(&to_string));
           if (v.compare("all") == 0 || v == str) return false;
         }
         return true;
@@ -1701,7 +1585,7 @@ namespace Sass {
         if (!l || !l->length()) return str.compare("rule") == 0;
         for (size_t i = 0, L = l->length(); i < L; ++i)
         {
-          v = unquote((*l)[i]->to_string());
+          v = unquote((*l)[i]->perform(&to_string));
           if (v.compare("all") == 0 || v == str) return true;
         }
         return false;
@@ -1765,6 +1649,7 @@ namespace Sass {
     }
 
     virtual bool operator== (const Expression& rhs) const;
+    virtual std::string to_string(bool compressed = false, int precision = 5) const;
 
     ATTACH_OPERATIONS()
   };
@@ -1846,7 +1731,7 @@ namespace Sass {
   // Abstract base class for CSS selectors.
   /////////////////////////////////////////
   class Selector : public Expression {
-    // ADD_PROPERTY(bool, has_reference)
+    ADD_PROPERTY(bool, has_reference)
     ADD_PROPERTY(bool, has_placeholder)
     // line break before list separator
     ADD_PROPERTY(bool, has_line_feed)
@@ -1856,24 +1741,17 @@ namespace Sass {
     ADD_PROPERTY(bool, is_optional)
     // parent block pointers
     ADD_PROPERTY(Media_Block*, media_block)
-  protected:
-    size_t hash_;
   public:
     Selector(ParserState pstate, bool r = false, bool h = false)
     : Expression(pstate),
-      // has_reference_(r),
+      has_reference_(r),
       has_placeholder_(h),
       has_line_feed_(false),
       has_line_break_(false),
       is_optional_(false),
-      media_block_(0),
-      hash_(0)
+      media_block_(0)
     { concrete_type(SELECTOR); }
     virtual ~Selector() = 0;
-    virtual size_t hash() = 0;
-    virtual bool has_parent_ref() {
-      return false;
-    }
     virtual unsigned long specificity() {
       return Constants::Specificity_Universal;
     }
@@ -1891,12 +1769,6 @@ namespace Sass {
     Selector_Schema(ParserState pstate, String* c)
     : Selector(pstate), contents_(c), at_root_(false)
     { }
-    virtual size_t hash() {
-      if (hash_ == 0) {
-        hash_combine(hash_, contents_->hash());
-      }
-      return hash_;
-    }
     ATTACH_OPERATIONS()
   };
 
@@ -1926,15 +1798,6 @@ namespace Sass {
         name += ns_ + "|";
       return name + name_;
     }
-    virtual size_t hash()
-    {
-      if (hash_ == 0) {
-        hash_combine(hash_, std::hash<int>()(SELECTOR));
-        hash_combine(hash_, std::hash<std::string>()(ns()));
-        hash_combine(hash_, std::hash<std::string>()(name()));
-      }
-      return hash_;
-    }
     // namespace query functions
     bool is_universal_ns() const
     {
@@ -1974,7 +1837,6 @@ namespace Sass {
     inline bool operator!=(const Simple_Selector& rhs) const { return !(*this == rhs); }
 
     bool operator<(const Simple_Selector& rhs) const;
-    // default implementation should work for most of the simple selectors (otherwise overload)
     ATTACH_OPERATIONS();
   };
   inline Simple_Selector::~Simple_Selector() { }
@@ -1990,7 +1852,7 @@ namespace Sass {
   public:
     Parent_Selector(ParserState pstate)
     : Simple_Selector(pstate, "&")
-    { /* has_reference(true); */ }
+    { has_reference(true); }
     virtual bool has_parent_ref() { return true; };
     virtual unsigned long specificity()
     {
@@ -2010,7 +1872,6 @@ namespace Sass {
     : Simple_Selector(pstate, n)
     { has_placeholder(true); }
     // virtual Selector_Placeholder* find_placeholder();
-    virtual ~Selector_Placeholder() {};
     ATTACH_OPERATIONS()
   };
 
@@ -2061,23 +1922,10 @@ namespace Sass {
     Attribute_Selector(ParserState pstate, std::string n, std::string m, String* v)
     : Simple_Selector(pstate, n), matcher_(m), value_(v)
     { }
-    virtual size_t hash()
-    {
-      if (hash_ == 0) {
-        hash_combine(hash_, Simple_Selector::hash());
-        hash_combine(hash_, std::hash<std::string>()(matcher()));
-        if (value_) hash_combine(hash_, value_->hash());
-      }
-      return hash_;
-    }
     virtual unsigned long specificity()
     {
       return Constants::Specificity_Attr;
     }
-    bool operator==(const Simple_Selector& rhs) const;
-    bool operator==(const Attribute_Selector& rhs) const;
-    bool operator<(const Simple_Selector& rhs) const;
-    bool operator<(const Attribute_Selector& rhs) const;
     ATTACH_OPERATIONS()
   };
 
@@ -2096,7 +1944,6 @@ namespace Sass {
            name == ":first-letter";
   }
 
-  // Pseudo Selector cannot have any namespace?
   class Pseudo_Selector : public Simple_Selector {
     ADD_PROPERTY(String*, expression)
   public:
@@ -2125,14 +1972,6 @@ namespace Sass {
       return (name_[0] == ':' && name_[1] == ':')
              || is_pseudo_class_element(name_);
     }
-    virtual size_t hash()
-    {
-      if (hash_ == 0) {
-        hash_combine(hash_, Simple_Selector::hash());
-        if (expression_) hash_combine(hash_, expression_->hash());
-      }
-      return hash_;
-    }
     virtual unsigned long specificity()
     {
       if (is_pseudo_element())
@@ -2152,30 +1991,15 @@ namespace Sass {
     Wrapped_Selector(ParserState pstate, std::string n, Selector* sel)
     : Simple_Selector(pstate, n), selector_(sel)
     { }
-    virtual bool has_parent_ref() {
-      // if (has_reference()) return true;
-      if (!selector()) return false;
-      return selector()->has_parent_ref();
-    }
     virtual bool is_superselector_of(Wrapped_Selector* sub);
     // Selectors inside the negation pseudo-class are counted like any
     // other, but the negation itself does not count as a pseudo-class.
-    virtual size_t hash()
-    {
-      if (hash_ == 0) {
-        hash_combine(hash_, Simple_Selector::hash());
-        if (selector_) hash_combine(hash_, selector_->hash());
-      }
-      return hash_;
-    }
     virtual unsigned long specificity()
     {
       return selector_ ? selector_->specificity() : 0;
     }
     bool operator==(const Simple_Selector& rhs) const;
     bool operator==(const Wrapped_Selector& rhs) const;
-    bool operator<(const Simple_Selector& rhs) const;
-    bool operator<(const Wrapped_Selector& rhs) const;
     ATTACH_OPERATIONS()
   };
 
@@ -2191,19 +2015,17 @@ namespace Sass {
   class Compound_Selector : public Selector, public Vectorized<Simple_Selector*> {
   private:
     SourcesSet sources_;
-    ADD_PROPERTY(bool, extended);
     ADD_PROPERTY(bool, has_parent_reference);
   protected:
     void adjust_after_pushing(Simple_Selector* s)
     {
-      // if (s->has_reference())   has_reference(true);
+      if (s->has_reference())   has_reference(true);
       if (s->has_placeholder()) has_placeholder(true);
     }
   public:
     Compound_Selector(ParserState pstate, size_t s = 0)
     : Selector(pstate),
       Vectorized<Simple_Selector*>(s),
-      extended_(false),
       has_parent_reference_(false)
     { }
     bool contains_placeholder() {
@@ -2218,7 +2040,7 @@ namespace Sass {
       return length() == 1 && (*this)[0]->is_universal();
     }
 
-    Complex_Selector* to_complex(Memory_Manager& mem);
+    Complex_Selector* to_complex(Memory_Manager<AST_Node>& mem);
     Compound_Selector* unify_with(Compound_Selector* rhs, Context& ctx);
     // virtual Selector_Placeholder* find_placeholder();
     virtual bool has_parent_ref();
@@ -2229,22 +2051,14 @@ namespace Sass {
     }
     const Simple_Selector* base() const {
       if (length() == 0) return 0;
-      // ToDo: why is this needed?
-      if (dynamic_cast<Type_Selector*>((*this)[0]))
+      if (typeid(*(*this)[0]) == typeid(Type_Selector))
         return (*this)[0];
+//      else cerr << "SERIOUSELY " << "\n";
       return 0;
     }
     virtual bool is_superselector_of(Compound_Selector* sub, std::string wrapped = "");
     virtual bool is_superselector_of(Complex_Selector* sub, std::string wrapped = "");
     virtual bool is_superselector_of(Selector_List* sub, std::string wrapped = "");
-    virtual size_t hash()
-    {
-      if (Selector::hash_ == 0) {
-        hash_combine(Selector::hash_, std::hash<int>()(SELECTOR));
-        if (length()) hash_combine(Selector::hash_, Vectorized::hash());
-      }
-      return Selector::hash_;
-    }
     virtual unsigned long specificity()
     {
       int sum = 0;
@@ -2256,7 +2070,7 @@ namespace Sass {
     bool is_empty_reference()
     {
       return length() == 1 &&
-             dynamic_cast<Parent_Selector*>((*this)[0]);
+             typeid(*(*this)[0]) == typeid(Parent_Selector);
     }
     std::vector<std::string> to_str_vec(); // sometimes need to convert to a flat "by-value" data structure
 
@@ -2299,12 +2113,9 @@ namespace Sass {
                      Compound_Selector* h = 0,
                      Complex_Selector* t = 0,
                      String* r = 0)
-    : Selector(pstate),
-      combinator_(c),
-      head_(h), tail_(t),
-      reference_(r)
+    : Selector(pstate), combinator_(c), head_(h), tail_(t), reference_(r)
     {
-      // if ((h && h->has_reference())   || (t && t->has_reference()))   has_reference(true);
+      if ((h && h->has_reference())   || (t && t->has_reference()))   has_reference(true);
       if ((h && h->has_placeholder()) || (t && t->has_placeholder())) has_placeholder(true);
     }
     virtual bool has_parent_ref();
@@ -2314,10 +2125,9 @@ namespace Sass {
       if ((!head_ || !head_->length() || head_->is_empty_reference()) &&
           combinator() == Combinator::ANCESTOR_OF)
       {
-        if (!tail_) return 0;
         tail_->has_line_feed_ = this->has_line_feed_;
         // tail_->has_line_break_ = this->has_line_break_;
-        return tail_->skip_empty_reference();
+        return tail_ ? tail_->skip_empty_reference() : 0;
       }
       return this;
     }
@@ -2359,16 +2169,6 @@ namespace Sass {
     Combinator clear_innermost();
     void append(Context&, Complex_Selector*);
     void set_innermost(Complex_Selector*, Combinator);
-    virtual size_t hash()
-    {
-      if (hash_ == 0) {
-        hash_combine(hash_, std::hash<int>()(SELECTOR));
-        hash_combine(hash_, std::hash<int>()(combinator_));
-        if (head_) hash_combine(hash_, head_->hash());
-        if (tail_) hash_combine(hash_, tail_->hash());
-      }
-      return hash_;
-    }
     virtual unsigned long specificity() const
     {
       int sum = 0;
@@ -2447,10 +2247,8 @@ namespace Sass {
     Selector_List(ParserState pstate, size_t s = 0)
     : Selector(pstate), Vectorized<Complex_Selector*>(s), wspace_(0)
     { }
-    std::string type() { return "list"; }
     // remove parent selector references
     // basically unwraps parsed selectors
-    virtual bool has_parent_ref();
     void remove_parent_selectors();
     // virtual Selector_Placeholder* find_placeholder();
     Selector_List* parentize(Selector_List* parents, Context& ctx);
@@ -2459,14 +2257,6 @@ namespace Sass {
     virtual bool is_superselector_of(Selector_List* sub, std::string wrapping = "");
     Selector_List* unify_with(Selector_List*, Context&);
     void populate_extends(Selector_List*, Context&, ExtensionSubsetMap&);
-    virtual size_t hash()
-    {
-      if (Selector::hash_ == 0) {
-        hash_combine(Selector::hash_, std::hash<int>()(SELECTOR));
-        hash_combine(Selector::hash_, Vectorized::hash());
-      }
-      return Selector::hash_;
-    }
     virtual unsigned long specificity()
     {
       unsigned long sum = 0;
@@ -2482,8 +2272,6 @@ namespace Sass {
     Selector_List* cloneFully(Context&) const; // clones Compound_Selector*s
     virtual bool operator==(const Selector& rhs) const;
     virtual bool operator==(const Selector_List& rhs) const;
-    // Selector Lists can be compared to comma lists
-    virtual bool operator==(const Expression& rhs) const;
     ATTACH_OPERATIONS()
   };
 
