@@ -2,8 +2,6 @@
 #define SASS_EMITTER_H
 
 #include <string>
-#include "sass.hpp"
-#include "sass/base.h"
 #include "source_map.hpp"
 #include "ast_fwd_decl.hpp"
 
@@ -13,7 +11,7 @@ namespace Sass {
   class Emitter {
 
     public:
-      Emitter(struct Sass_Output_Options& opt);
+      Emitter(Context* ctx);
       virtual ~Emitter() { }
 
     protected:
@@ -25,19 +23,17 @@ namespace Sass {
       // proxy methods for source maps
       void add_source_index(size_t idx);
       void set_filename(const std::string& str);
-      void add_open_mapping(const AST_Node* node);
-      void add_close_mapping(const AST_Node* node);
-      void schedule_mapping(const AST_Node* node);
-      std::string render_srcmap(Context &ctx);
+      void add_open_mapping(AST_Node* node);
+      void add_close_mapping(AST_Node* node);
+      std::string generate_source_map(Context &ctx);
       ParserState remap(const ParserState& pstate);
 
     public:
-      struct Sass_Output_Options& opt;
+      Context* ctx;
       size_t indentation;
       size_t scheduled_space;
       size_t scheduled_linefeed;
       bool scheduled_delimiter;
-      const AST_Node* scheduled_mapping;
 
     public:
       // output strings different in comments
@@ -56,9 +52,9 @@ namespace Sass {
       // return buffer as std::string
       std::string get_buffer(void);
       // flush scheduled space/linefeed
-      Sass_Output_Style output_style(void) const;
+      Output_Style output_style(void);
       // add outstanding linefeed
-      void finalize(bool final = true);
+      void finalize(void);
       // flush scheduled space/linefeed
       void flush_schedules(void);
       // prepend some text or token to the buffer
@@ -70,7 +66,7 @@ namespace Sass {
       void append_wspace(const std::string& text);
       // append some text or token to the buffer
       // this adds source-mappings for node start and end
-      void append_token(const std::string& text, const AST_Node* node);
+      void append_token(const std::string& text, AST_Node* node);
 
     public: // syntax sugar
       void append_indentation();
