@@ -1,20 +1,19 @@
-#include "sass.hpp"
 #include "functions.hpp"
 #include "ast.hpp"
 #include "context.hpp"
 #include "backtrace.hpp"
 #include "parser.hpp"
 #include "constants.hpp"
+#include "to_string.hpp"
 #include "inspect.hpp"
 #include "extend.hpp"
 #include "eval.hpp"
 #include "util.hpp"
 #include "expand.hpp"
 #include "utf8_string.hpp"
-#include "sass/base.h"
 #include "utf8.h"
 
-#include <cstdint>
+#include <atomic>
 #include <cstdlib>
 #include <cmath>
 #include <cctype>
@@ -156,6 +155,7 @@ namespace Sass {
 
     template <>
     Selector_List* get_arg_sel(const std::string& argname, Env& env, Signature sig, ParserState pstate, Backtrace* backtrace, Context& ctx) {
+      To_String to_string(&ctx, false);
       Expression* exp = ARG(argname, Expression);
       if (exp->concrete_type() == Expression::NULL_VAL) {
         std::stringstream msg;
@@ -163,15 +163,13 @@ namespace Sass {
         msg << "a list of strings, or a list of lists of strings for `" << function_name(sig) << "'";
         error(msg.str(), pstate);
       }
-      if (String_Constant* str =dynamic_cast<String_Constant*>(exp)) {
-        str->quote_mark(0);
-      }
-      std::string exp_src = exp->to_string(ctx.c_options) + "{";
+      std::string exp_src = exp->perform(&to_string) + "{";
       return Parser::parse_selector(exp_src.c_str(), ctx);
     }
 
     template <>
     Complex_Selector* get_arg_sel(const std::string& argname, Env& env, Signature sig, ParserState pstate, Backtrace* backtrace, Context& ctx) {
+      To_String to_string(&ctx, false);
       Expression* exp = ARG(argname, Expression);
       if (exp->concrete_type() == Expression::NULL_VAL) {
         std::stringstream msg;
@@ -179,26 +177,22 @@ namespace Sass {
         msg << "a list of strings, or a list of lists of strings for `" << function_name(sig) << "'";
         error(msg.str(), pstate);
       }
-      if (String_Constant* str =dynamic_cast<String_Constant*>(exp)) {
-        str->quote_mark(0);
-      }
-      std::string exp_src = exp->to_string(ctx.c_options) + "{";
+      std::string exp_src = exp->perform(&to_string) + "{";
       Selector_List* sel_list = Parser::parse_selector(exp_src.c_str(), ctx);
       return (sel_list->length() > 0) ? sel_list->first() : 0;
     }
 
     template <>
     Compound_Selector* get_arg_sel(const std::string& argname, Env& env, Signature sig, ParserState pstate, Backtrace* backtrace, Context& ctx) {
+      To_String to_string(&ctx, false);
       Expression* exp = ARG(argname, Expression);
       if (exp->concrete_type() == Expression::NULL_VAL) {
         std::stringstream msg;
-        msg << argname << ": null is not a string for `" << function_name(sig) << "'";
+        msg << argname << ": null is not a valid selector: it must be a string,\n";
+        msg << "a list of strings, or a list of lists of strings for `" << function_name(sig) << "'";
         error(msg.str(), pstate);
       }
-      if (String_Constant* str =dynamic_cast<String_Constant*>(exp)) {
-        str->quote_mark(0);
-      }
-      std::string exp_src = exp->to_string(ctx.c_options) + "{";
+      std::string exp_src = exp->perform(&to_string) + "{";
       Selector_List* sel_list = Parser::parse_selector(exp_src.c_str(), ctx);
       return (sel_list->length() > 0) ? sel_list->first()->tail()->head() : 0;
     }
@@ -218,9 +212,9 @@ namespace Sass {
       return seed;
     }
     #else
+    static std::random_device rd;
     uint64_t GetSeed()
     {
-      std::random_device rd;
       return rd();
     }
     #endif
@@ -318,9 +312,9 @@ namespace Sass {
 
       return SASS_MEMORY_NEW(ctx.mem, Color,
                              pstate,
-                             Sass::round(w1*color1->r() + w2*color2->r(), ctx.c_options.precision),
-                             Sass::round(w1*color1->g() + w2*color2->g(), ctx.c_options.precision),
-                             Sass::round(w1*color1->b() + w2*color2->b(), ctx.c_options.precision),
+                             std::round(w1*color1->r() + w2*color2->r()),
+                             std::round(w1*color1->g() + w2*color2->g()),
+                             std::round(w1*color1->b() + w2*color2->b()),
                              color1->a()*p + color2->a()*(1-p));
     }
 
@@ -338,7 +332,7 @@ namespace Sass {
 
       double max = std::max(r, std::max(g, b));
       double min = std::min(r, std::min(g, b));
-      double delta = max - min;
+      double del = max - min;
 
       double h = 0, s = 0, l = (max + min) / 2.0;
 
@@ -346,12 +340,12 @@ namespace Sass {
         h = s = 0; // achromatic
       }
       else {
-        if (l < 0.5) s = delta / (max + min);
-        else         s = delta / (2.0 - max - min);
+        if (l < 0.5) s = del / (max + min);
+        else         s = del / (2.0 - max - min);
 
-        if      (r == max) h = (g - b) / delta + (g < b ? 6 : 0);
-        else if (g == max) h = (b - r) / delta + 2;
-        else if (b == max) h = (r - g) / delta + 4;
+        if      (r == max) h = (g - b) / del + (g < b ? 6 : 0);
+        else if (g == max) h = (b - r) / del + 2;
+        else if (b == max) h = (r - g) / del + 4;
       }
 
       HSL hsl_struct;
@@ -364,8 +358,8 @@ namespace Sass {
 
     // hue to RGB helper function
     double h_to_rgb(double m1, double m2, double h) {
-      while (h < 0) h += 1;
-      while (h > 1) h -= 1;
+      if (h < 0) h += 1;
+      if (h > 1) h -= 1;
       if (h*6.0 < 1) return m1 + (m2 - m1)*h*6;
       if (h*2.0 < 1) return m2;
       if (h*3.0 < 2) return m1 + (m2 - m1) * (2.0/3.0 - h)*6;
@@ -391,9 +385,9 @@ namespace Sass {
       else m2 = (l+s)-(l*s);
       double m1 = (l*2.0)-m2;
       // round the results -- consider moving this into the Color constructor
-      double r = (h_to_rgb(m1, m2, h + 1.0/3.0) * 255.0);
+      double r = (h_to_rgb(m1, m2, h+1.0/3.0) * 255.0);
       double g = (h_to_rgb(m1, m2, h) * 255.0);
-      double b = (h_to_rgb(m1, m2, h - 1.0/3.0) * 255.0);
+      double b = (h_to_rgb(m1, m2, h-1.0/3.0) * 255.0);
 
       return SASS_MEMORY_NEW(ctx.mem, Color, pstate, r, g, b, a);
     }
@@ -517,7 +511,8 @@ namespace Sass {
       // CSS3 filter function overload: pass literal through directly
       Number* amount = dynamic_cast<Number*>(env["$amount"]);
       if (!amount) {
-        return SASS_MEMORY_NEW(ctx.mem, String_Quoted, pstate, "saturate(" + env["$color"]->to_string(ctx.c_options) + ")");
+        To_String to_string(&ctx);
+        return SASS_MEMORY_NEW(ctx.mem, String_Quoted, pstate, "saturate(" + env["$color"]->perform(&to_string) + ")");
       }
 
       ARGR("$amount", Number, 0, 100);
@@ -577,7 +572,8 @@ namespace Sass {
       // CSS3 filter function overload: pass literal through directly
       Number* amount = dynamic_cast<Number*>(env["$color"]);
       if (amount) {
-        return SASS_MEMORY_NEW(ctx.mem, String_Quoted, pstate, "grayscale(" + amount->to_string(ctx.c_options) + ")");
+        To_String to_string(&ctx);
+        return SASS_MEMORY_NEW(ctx.mem, String_Quoted, pstate, "grayscale(" + amount->perform(&to_string) + ")");
       }
 
       Color* rgb_color = ARG("$color", Color);
@@ -613,7 +609,8 @@ namespace Sass {
       // CSS3 filter function overload: pass literal through directly
       Number* amount = dynamic_cast<Number*>(env["$color"]);
       if (amount) {
-        return SASS_MEMORY_NEW(ctx.mem, String_Quoted, pstate, "invert(" + amount->to_string(ctx.c_options) + ")");
+        To_String to_string(&ctx);
+        return SASS_MEMORY_NEW(ctx.mem, String_Quoted, pstate, "invert(" + amount->perform(&to_string) + ")");
       }
 
       Color* rgb_color = ARG("$color", Color);
@@ -640,7 +637,8 @@ namespace Sass {
       // CSS3 filter function overload: pass literal through directly
       Number* amount = dynamic_cast<Number*>(env["$color"]);
       if (amount) {
-        return SASS_MEMORY_NEW(ctx.mem, String_Quoted, pstate, "opacity(" + amount->to_string(ctx.c_options) + ")");
+        To_String to_string(&ctx);
+        return SASS_MEMORY_NEW(ctx.mem, String_Quoted, pstate, "opacity(" + amount->perform(&to_string) + ")");
       }
 
       return SASS_MEMORY_NEW(ctx.mem, Number, pstate, ARG("$color", Color)->a());
@@ -857,10 +855,10 @@ namespace Sass {
 
       std::stringstream ss;
       ss << '#' << std::setw(2) << std::setfill('0');
-      ss << std::hex << std::setw(2) << static_cast<unsigned long>(Sass::round(a, ctx.c_options.precision));
-      ss << std::hex << std::setw(2) << static_cast<unsigned long>(Sass::round(r, ctx.c_options.precision));
-      ss << std::hex << std::setw(2) << static_cast<unsigned long>(Sass::round(g, ctx.c_options.precision));
-      ss << std::hex << std::setw(2) << static_cast<unsigned long>(Sass::round(b, ctx.c_options.precision));
+      ss << std::hex << std::setw(2) << static_cast<unsigned long>(std::floor(a+0.5));
+      ss << std::hex << std::setw(2) << static_cast<unsigned long>(std::floor(r+0.5));
+      ss << std::hex << std::setw(2) << static_cast<unsigned long>(std::floor(g+0.5));
+      ss << std::hex << std::setw(2) << static_cast<unsigned long>(std::floor(b+0.5));
 
       std::string result(ss.str());
       for (size_t i = 0, L = result.length(); i < L; ++i) {
@@ -877,7 +875,10 @@ namespace Sass {
     BUILT_IN(sass_unquote)
     {
       AST_Node* arg = env["$string"];
-      if (String_Quoted* string_quoted = dynamic_cast<String_Quoted*>(arg)) {
+      if (dynamic_cast<Null*>(arg)) {
+        return SASS_MEMORY_NEW(ctx.mem, Null, pstate);
+      }
+      else if (String_Quoted* string_quoted = dynamic_cast<String_Quoted*>(arg)) {
         String_Constant* result = SASS_MEMORY_NEW(ctx.mem, String_Constant, pstate, string_quoted->value());
         // remember if the string was quoted (color tokens)
         result->sass_fix_1291(string_quoted->quote_mark() != 0);
@@ -887,13 +888,9 @@ namespace Sass {
         return (Expression*) arg;
       }
       else {
-        Sass_Output_Style oldstyle = ctx.c_options.output_style;
-        ctx.c_options.output_style = SASS_STYLE_NESTED;
-        std::string val(arg->to_string(ctx.c_options));
-        val = dynamic_cast<Null*>(arg) ? "null" : val;
-        ctx.c_options.output_style = oldstyle;
-
-        deprecated_function("Passing " + val + ", a non-string value, to unquote()", pstate);
+        To_String to_string(&ctx);
+        std::string val(arg->perform(&to_string));
+        deprecated("Passing " + val + ", a non-string value, to unquote()", pstate);
         return (Expression*) arg;
       }
     }
@@ -901,8 +898,9 @@ namespace Sass {
     Signature quote_sig = "quote($string)";
     BUILT_IN(sass_quote)
     {
+      To_String to_string(&ctx);
       AST_Node* arg = env["$string"];
-      std::string str(quote(arg->to_string(ctx.c_options), String_Constant::double_quote()));
+      std::string str(quote(arg->perform(&to_string), String_Constant::double_quote()));
       String_Quoted* result = SASS_MEMORY_NEW(ctx.mem, String_Quoted, pstate, str);
       result->is_delayed(true);
       return result;
@@ -1092,7 +1090,7 @@ namespace Sass {
       Number* n = ARG("$number", Number);
       Number* r = SASS_MEMORY_NEW(ctx.mem, Number, *n);
       r->pstate(pstate);
-      r->value(Sass::round(r->value(), ctx.c_options.precision));
+      r->value(std::floor(r->value() + 0.5));
       return r;
     }
 
@@ -1132,13 +1130,10 @@ namespace Sass {
       List* arglist = ARG("$numbers", List);
       Number* least = 0;
       for (size_t i = 0, L = arglist->length(); i < L; ++i) {
-        Expression* val = arglist->value_at_index(i);
-        Number* xi = dynamic_cast<Number*>(val);
-        if (!xi) {
-          error("\"" + val->to_string(ctx.c_options) + "\" is not a number for `min'", pstate);
-        }
+        Number* xi = dynamic_cast<Number*>(arglist->value_at_index(i));
+        if (!xi) error("`" + std::string(sig) + "` only takes numeric arguments", pstate);
         if (least) {
-          if (*xi < *least) least = xi;
+          if (Eval::lt(xi, least)) least = xi;
         } else least = xi;
       }
       return least;
@@ -1150,13 +1145,10 @@ namespace Sass {
       List* arglist = ARG("$numbers", List);
       Number* greatest = 0;
       for (size_t i = 0, L = arglist->length(); i < L; ++i) {
-        Expression* val = arglist->value_at_index(i);
-        Number* xi = dynamic_cast<Number*>(val);
-        if (!xi) {
-          error("\"" + val->to_string(ctx.c_options) + "\" is not a number for `max'", pstate);
-        }
+        Number* xi = dynamic_cast<Number*>(arglist->value_at_index(i));
+        if (!xi) error("`" + std::string(sig) + "` only takes numeric arguments", pstate);
         if (greatest) {
-          if (*greatest < *xi) greatest = xi;
+          if (Eval::lt(greatest, xi)) greatest = xi;
         } else greatest = xi;
       }
       return greatest;
@@ -1165,10 +1157,7 @@ namespace Sass {
     Signature random_sig = "random($limit:false)";
     BUILT_IN(random)
     {
-      AST_Node* arg = env["$limit"];
-      Value* v = dynamic_cast<Value*>(arg);
-      Number* l = dynamic_cast<Number*>(arg);
-      Boolean* b = dynamic_cast<Boolean*>(arg);
+      Number* l = dynamic_cast<Number*>(env["$limit"]);
       if (l) {
         double v = l->value();
         if (v < 1) {
@@ -1186,16 +1175,11 @@ namespace Sass {
         uint_fast32_t distributed = static_cast<uint_fast32_t>(distributor(rand));
         return SASS_MEMORY_NEW(ctx.mem, Number, pstate, (double)distributed);
       }
-      else if (b) {
+      else {
         std::uniform_real_distribution<> distributor(0, 1);
         double distributed = static_cast<double>(distributor(rand));
         return SASS_MEMORY_NEW(ctx.mem, Number, pstate, distributed);
-      } else if (v) {
-        throw Exception::InvalidArgumentType(pstate, "random", "$limit", "number", v);
-      } else {
-        throw Exception::InvalidArgumentType(pstate, "random", "$limit", "number");
-      }
-      return 0;
+     }
     }
 
     /////////////////
@@ -1205,13 +1189,12 @@ namespace Sass {
     Signature length_sig = "length($list)";
     BUILT_IN(length)
     {
-      if (Selector_List* sl = dynamic_cast<Selector_List*>(env["$list"])) {
-        return SASS_MEMORY_NEW(ctx.mem, Number, pstate, (double)sl->length());
-      }
       Expression* v = ARG("$list", Expression);
       if (v->concrete_type() == Expression::MAP) {
         Map* map = dynamic_cast<Map*>(env["$list"]);
-        return SASS_MEMORY_NEW(ctx.mem, Number, pstate, (double)(map ? map->length() : 1));
+        return SASS_MEMORY_NEW(ctx.mem, Number,
+     pstate,
+     (double)(map ? map->length() : 1));
       }
       if (v->concrete_type() == Expression::SELECTOR) {
         if (Compound_Selector* h = dynamic_cast<Compound_Selector*>(v)) {
@@ -1232,19 +1215,9 @@ namespace Sass {
     Signature nth_sig = "nth($list, $n)";
     BUILT_IN(nth)
     {
-      Number* n = ARG("$n", Number);
       Map* m = dynamic_cast<Map*>(env["$list"]);
-      if (Selector_List* sl = dynamic_cast<Selector_List*>(env["$list"])) {
-        size_t len = m ? m->length() : sl->length();
-        bool empty = m ? m->empty() : sl->empty();
-        if (empty) error("argument `$list` of `" + std::string(sig) + "` must not be empty", pstate);
-        double index = std::floor(n->value() < 0 ? len + n->value() : n->value() - 1);
-        if (index < 0 || index > len - 1) error("index out of bounds for `" + std::string(sig) + "`", pstate);
-        // return (*sl)[static_cast<int>(index)];
-        Listize listize(ctx.mem);
-        return (*sl)[static_cast<int>(index)]->perform(&listize);
-      }
       List* l = dynamic_cast<List*>(env["$list"]);
+      Number* n = ARG("$n", Number);
       if (n->value() == 0) error("argument `$n` of `" + std::string(sig) + "` must be non-zero", pstate);
       // if the argument isn't a list, then wrap it in a singleton list
       if (!m && !l) {
@@ -1335,10 +1308,6 @@ namespace Sass {
     {
       List* l = dynamic_cast<List*>(env["$list"]);
       Expression* v = ARG("$val", Expression);
-      if (Selector_List* sl = dynamic_cast<Selector_List*>(env["$list"])) {
-        Listize listize(ctx.mem);
-        l = dynamic_cast<List*>(sl->perform(&listize));
-      }
       String_Constant* sep = ARG("$separator", String_Constant);
       if (!l) {
         l = SASS_MEMORY_NEW(ctx.mem, List, pstate, 1);
@@ -1605,23 +1574,22 @@ namespace Sass {
       List* arglist = SASS_MEMORY_NEW(ctx.mem, List, *ARG("$args", List));
 
       Arguments* args = SASS_MEMORY_NEW(ctx.mem, Arguments, pstate);
-      // std::string full_name(name + "[f]");
-      // Definition* def = d_env.has(full_name) ? static_cast<Definition*>((d_env)[full_name]) : 0;
-      // Parameters* params = def ? def->parameters() : 0;
-      // size_t param_size = params ? params->length() : 0;
+      std::string full_name(name + "[f]");
+      Definition* def = d_env.has(full_name) ? static_cast<Definition*>((d_env)[full_name]) : 0;
+      Parameters* params = def ? def->parameters() : 0;
+      size_t param_size = params ? params->length() : 0;
       for (size_t i = 0, L = arglist->length(); i < L; ++i) {
         Expression* expr = arglist->value_at_index(i);
-        // if (params && params->has_rest_parameter()) {
-        //   Parameter* p = param_size > i ? (*params)[i] : 0;
-        //   List* list = dynamic_cast<List*>(expr);
-        //   if (list && p && !p->is_rest_parameter()) expr = (*list)[0];
-        // }
+        Parameter* p = param_size > i ? (*params)[i] : 0;
+        if (List* list = dynamic_cast<List*>(expr)) {
+          if (p && !p->is_rest_parameter()) expr = (*list)[0];
+        }
         if (arglist->is_arglist()) {
           Argument* arg = dynamic_cast<Argument*>((*arglist)[i]);
           *args << SASS_MEMORY_NEW(ctx.mem, Argument,
                                    pstate,
                                    expr,
-                                   arg ? arg->name() : "",
+                                   "",
                                    arg ? arg->is_rest_argument() : false,
                                    arg ? arg->is_keyword_argument() : false);
         } else {
@@ -1672,9 +1640,6 @@ namespace Sass {
     // MISCELLANEOUS FUNCTIONS
     //////////////////////////
 
-    // value.check_deprecated_interp if value.is_a?(Sass::Script::Value::String)
-    // unquoted_string(value.to_sass)
-
     Signature inspect_sig = "inspect($value)";
     BUILT_IN(inspect)
     {
@@ -1686,23 +1651,23 @@ namespace Sass {
       } else if (v->concrete_type() == Expression::STRING) {
         return v;
       } else {
-        // ToDo: fix to_sass for nested parentheses
-        Sass_Output_Style old_style;
-        old_style = ctx.c_options.output_style;
-        ctx.c_options.output_style = TO_SASS;
-        Sass_Output_Options out(ctx.c_options);
-        Emitter emitter(ctx.c_options);
-        Inspect i(emitter);
-        i.in_declaration = false;
-        v->perform(&i);
-        ctx.c_options.output_style = old_style;
-        return SASS_MEMORY_NEW(ctx.mem, String_Quoted, pstate, i.get_buffer());
+        bool parentheses = v->concrete_type() == Expression::MAP ||
+                           v->concrete_type() == Expression::LIST;
+        Output_Style old_style;
+        old_style = ctx.output_style;
+        ctx.output_style = NESTED;
+        To_String to_string(&ctx, false);
+        std::string inspect = v->perform(&to_string);
+        if (inspect.empty() && parentheses) inspect = "()";
+        ctx.output_style = old_style;
+        return SASS_MEMORY_NEW(ctx.mem, String_Quoted, pstate, inspect);
       }
       // return v;
     }
     Signature selector_nest_sig = "selector-nest($selectors...)";
     BUILT_IN(selector_nest)
     {
+      To_String to_string(&ctx, false);
       List* arglist = ARG("$selectors", List);
 
       // Not enough parameters
@@ -1719,10 +1684,7 @@ namespace Sass {
           msg << "a list of strings, or a list of lists of strings for 'selector-nest'";
           error(msg.str(), pstate);
         }
-        if (String_Constant* str =dynamic_cast<String_Constant*>(exp)) {
-          str->quote_mark(0);
-        }
-        std::string exp_src = exp->to_string(ctx.c_options) + "{";
+        std::string exp_src = exp->perform(&to_string) + "{";
         Selector_List* sel = Parser::parse_selector(exp_src.c_str(), ctx);
         parsedSelectors.push_back(sel);
       }
@@ -1747,13 +1709,14 @@ namespace Sass {
         result->elements(exploded);
       }
 
-      Listize listize(ctx.mem);
+      Listize listize(ctx);
       return result->perform(&listize);
     }
 
     Signature selector_append_sig = "selector-append($selectors...)";
     BUILT_IN(selector_append)
     {
+      To_String to_string;
       List* arglist = ARG("$selectors", List);
 
       // Not enough parameters
@@ -1770,10 +1733,7 @@ namespace Sass {
           msg << "a list of strings, or a list of lists of strings for 'selector-append'";
           error(msg.str(), pstate);
         }
-        if (String_Constant* str =dynamic_cast<String_Constant*>(exp)) {
-          str->quote_mark(0);
-        }
-        std::string exp_src = exp->to_string() + "{";
+        std::string exp_src = exp->perform(&to_string) + "{";
         Selector_List* sel = Parser::parse_selector(exp_src.c_str(), ctx);
         parsedSelectors.push_back(sel);
       }
@@ -1808,9 +1768,9 @@ namespace Sass {
             // Must be a simple sequence
             if( childSeq->combinator() != Complex_Selector::Combinator::ANCESTOR_OF ) {
               std::string msg("Can't append  `");
-              msg += childSeq->to_string();
+              msg += childSeq->perform(&to_string);
               msg += "` to `";
-              msg += parentSeqClone->to_string();
+              msg += parentSeqClone->perform(&to_string);;
               msg += "`";
               error(msg, pstate, backtrace);
             }
@@ -1819,9 +1779,9 @@ namespace Sass {
             Type_Selector* pType = dynamic_cast<Type_Selector*>(base->head()->first());
             if(pType && pType->name() == "*") {
               std::string msg("Can't append  `");
-              msg += childSeq->to_string();
+              msg += childSeq->perform(&to_string);
               msg += "` to `";
-              msg += parentSeqClone->to_string();
+              msg += parentSeqClone->perform(&to_string);;
               msg += "`";
               error(msg, pstate, backtrace);
             }
@@ -1841,7 +1801,7 @@ namespace Sass {
         result->elements(newElements);
       }
 
-      Listize listize(ctx.mem);
+      Listize listize(ctx);
       return result->perform(&listize);
     }
 
@@ -1852,7 +1812,7 @@ namespace Sass {
       Selector_List* selector2 = ARGSEL("$selector2", Selector_List, p_contextualize);
 
       Selector_List* result = selector1->unify_with(selector2, ctx);
-      Listize listize(ctx.mem);
+      Listize listize(ctx);
       return result->perform(&listize);
     }
 
@@ -1861,11 +1821,12 @@ namespace Sass {
     {
       Compound_Selector* sel = ARGSEL("$selector", Compound_Selector, p_contextualize);
 
+      To_String to_string;
       List* l = SASS_MEMORY_NEW(ctx.mem, List, sel->pstate(), sel->length(), SASS_COMMA);
 
       for (size_t i = 0, L = sel->length(); i < L; ++i) {
         Simple_Selector* ss = (*sel)[i];
-        std::string ss_string = ss->to_string() ;
+        std::string ss_string = ss->perform(&to_string) ;
 
         *l << SASS_MEMORY_NEW(ctx.mem, String_Quoted, ss->pstate(), ss_string);
       }
@@ -1886,7 +1847,7 @@ namespace Sass {
       bool extendedSomething;
       Selector_List* result = Extend::extendSelectorList(selector, ctx, subset_map, false, extendedSomething);
 
-      Listize listize(ctx.mem);
+      Listize listize(ctx);
       return result->perform(&listize);
     }
 
@@ -1903,7 +1864,7 @@ namespace Sass {
       bool extendedSomething;
       Selector_List* result = Extend::extendSelectorList(selector, ctx, subset_map, true, extendedSomething);
 
-      Listize listize(ctx.mem);
+      Listize listize(ctx);
       return result->perform(&listize);
     }
 
@@ -1912,15 +1873,20 @@ namespace Sass {
     {
       Selector_List* sel = ARGSEL("$selector", Selector_List, p_contextualize);
 
-      Listize listize(ctx.mem);
+      Listize listize(ctx);
       return sel->perform(&listize);
     }
 
     Signature is_superselector_sig = "is-superselector($super, $sub)";
     BUILT_IN(is_superselector)
     {
-      Selector_List*  sel_sup = ARGSEL("$super", Selector_List, p_contextualize);
-      Selector_List*  sel_sub = ARGSEL("$sub", Selector_List, p_contextualize);
+      To_String to_string(&ctx, false);
+      Expression*  ex_sup = ARG("$super", Expression);
+      Expression*  ex_sub = ARG("$sub", Expression);
+      std::string sup_src = ex_sup->perform(&to_string) + "{";
+      std::string sub_src = ex_sub->perform(&to_string) + "{";
+      Selector_List* sel_sup = Parser::parse_selector(sup_src.c_str(), ctx);
+      Selector_List* sel_sub = Parser::parse_selector(sub_src.c_str(), ctx);
       bool result = sel_sup->is_superselector_of(sel_sub);
       return SASS_MEMORY_NEW(ctx.mem, Boolean, pstate, result);
     }
