@@ -1,4 +1,3 @@
-#include "sass.hpp"
 #include <cmath>
 #include <string>
 #include <iostream>
@@ -131,8 +130,7 @@ namespace Sass {
     append_colon_separator();
 
     if (dec->value()->concrete_type() == Expression::SELECTOR) {
-      Memory_Manager mem;
-      Listize listize(mem);
+      Listize listize(*ctx);
       dec->value()->perform(&listize)->perform(this);
     } else {
       dec->value()->perform(this);
@@ -171,11 +169,9 @@ namespace Sass {
       }
 
       import->urls().front()->perform(this);
-      if (import->urls().size() == 1) {
-        if (import->media_queries()) {
-          append_mandatory_space();
-          import->media_queries()->perform(this);
-        }
+      if (import->media_queries()) {
+        append_mandatory_space();
+        import->media_queries()->perform(this);
       }
       append_delimiter();
       for (size_t i = 1, S = import->urls().size(); i < S; ++i) {
@@ -188,11 +184,9 @@ namespace Sass {
         }
 
         import->urls()[i]->perform(this);
-        if (import->urls().size() - 1 == i) {
-          if (import->media_queries()) {
-            append_mandatory_space();
-            import->media_queries()->perform(this);
-          }
+        if (import->media_queries()) {
+          append_mandatory_space();
+          import->media_queries()->perform(this);
         }
         append_delimiter();
       }
@@ -204,7 +198,7 @@ namespace Sass {
     append_indentation();
     append_token("@import", import);
     append_mandatory_space();
-    append_string(import->imp_path());
+    append_string(import->file_name());
     append_delimiter();
   }
 
@@ -352,10 +346,6 @@ namespace Sass {
 
   void Inspect::operator()(Map* map)
   {
-    if (output_style() == TO_SASS && map->empty()) {
-      append_string("()");
-      return;
-    }
     if (map->empty()) return;
     if (map->is_invisible()) return;
     bool items_output = false;
@@ -372,25 +362,15 @@ namespace Sass {
 
   void Inspect::operator()(List* list)
   {
-    if (output_style() == TO_SASS && list->empty()) {
-      append_string("()");
-      return;
-    }
     std::string sep(list->separator() == SASS_SPACE ? " " : ",");
-    if ((output_style() != COMPRESSED) && sep == ",") sep += " ";
+    if (output_style() != COMPRESSED && sep == ",") sep += " ";
     else if (in_media_block && sep != " ") sep += " "; // verified
     if (list->empty()) return;
     bool items_output = false;
 
     bool was_space_array = in_space_array;
     bool was_comma_array = in_comma_array;
-    // probably ruby sass eqivalent of element_needs_parens
-    if (output_style() == TO_SASS && list->length() == 1 &&
-      (!dynamic_cast<List*>((*list)[0]) &&
-       !dynamic_cast<Selector_List*>((*list)[0]))) {
-      append_string("(");
-    }
-    else if (!in_declaration && (list->separator() == SASS_HASH ||
+    if (!in_declaration && (
         (list->separator() == SASS_SPACE && in_space_array) ||
         (list->separator() == SASS_COMMA && in_comma_array)
     )) {
@@ -401,13 +381,9 @@ namespace Sass {
     else if (list->separator() == SASS_COMMA) in_comma_array = true;
 
     for (size_t i = 0, L = list->size(); i < L; ++i) {
-      if (list->separator() == SASS_HASH)
-      { sep[0] = i % 2 ? ':' : ','; }
       Expression* list_item = (*list)[i];
-      if (output_style() != TO_SASS) {
-        if (list_item->is_invisible()) {
-          continue;
-        }
+      if (list_item->is_invisible()) {
+        continue;
       }
       if (items_output) {
         append_string(sep);
@@ -420,13 +396,7 @@ namespace Sass {
 
     in_comma_array = was_comma_array;
     in_space_array = was_space_array;
-    // probably ruby sass eqivalent of element_needs_parens
-    if (output_style() == TO_SASS && list->length() == 1 &&
-      (!dynamic_cast<List*>((*list)[0]) &&
-       !dynamic_cast<Selector_List*>((*list)[0]))) {
-      append_string(",)");
-    }
-    else if (!in_declaration && (list->separator() == SASS_HASH ||
+    if (!in_declaration && (
         (list->separator() == SASS_SPACE && in_space_array) ||
         (list->separator() == SASS_COMMA && in_comma_array)
     )) {
@@ -438,41 +408,22 @@ namespace Sass {
   void Inspect::operator()(Binary_Expression* expr)
   {
     expr->left()->perform(this);
-    if ( in_media_block ||
-         (output_style() == INSPECT) || (
-          expr->op().ws_before
-          && (!expr->is_interpolant())
-          && (!expr->is_delayed() ||
-          expr->is_left_interpolant() ||
-          expr->is_right_interpolant()
-          )
-
-    )) append_string(" ");
     switch (expr->type()) {
-      case Sass_OP::AND: append_string("&&"); break;
-      case Sass_OP::OR:  append_string("||");  break;
-      case Sass_OP::EQ:  append_string("==");  break;
-      case Sass_OP::NEQ: append_string("!=");  break;
-      case Sass_OP::GT:  append_string(">");   break;
-      case Sass_OP::GTE: append_string(">=");  break;
-      case Sass_OP::LT:  append_string("<");   break;
-      case Sass_OP::LTE: append_string("<=");  break;
-      case Sass_OP::ADD: append_string("+");   break;
-      case Sass_OP::SUB: append_string("-");   break;
-      case Sass_OP::MUL: append_string("*");   break;
-      case Sass_OP::DIV: append_string("/"); break;
-      case Sass_OP::MOD: append_string("%");   break;
+      case Sass_OP::AND: append_string(" and "); break;
+      case Sass_OP::OR:  append_string(" or ");  break;
+      case Sass_OP::EQ:  append_string(" == ");  break;
+      case Sass_OP::NEQ: append_string(" != ");  break;
+      case Sass_OP::GT:  append_string(" > ");   break;
+      case Sass_OP::GTE: append_string(" >= ");  break;
+      case Sass_OP::LT:  append_string(" < ");   break;
+      case Sass_OP::LTE: append_string(" <= ");  break;
+      case Sass_OP::ADD: append_string(" + ");   break;
+      case Sass_OP::SUB: append_string(" - ");   break;
+      case Sass_OP::MUL: append_string(" * ");   break;
+      case Sass_OP::DIV: append_string(in_media_block ? " / " : "/"); break;
+      case Sass_OP::MOD: append_string(" % ");   break;
       default: break; // shouldn't get here
     }
-    if ( in_media_block ||
-         (output_style() == INSPECT) || (
-          expr->op().ws_after
-          && (!expr->is_interpolant())
-          && (!expr->is_delayed()
-              || expr->is_left_interpolant()
-              || expr->is_right_interpolant()
-          )
-    )) append_string(" ");
     expr->right()->perform(this);
   }
 
@@ -507,187 +458,29 @@ namespace Sass {
 
   void Inspect::operator()(Number* n)
   {
-
-    std::string res;
-
-    // check if the fractional part of the value equals to zero
-    // neat trick from http://stackoverflow.com/a/1521682/1550314
-    // double int_part; bool is_int = modf(value, &int_part) == 0.0;
-
-    // this all cannot be done with one run only, since fixed
-    // output differs from normal output and regular output
-    // can contain scientific notation which we do not want!
-
-    // first sample
-    std::stringstream ss;
-    ss.precision(12);
-    ss << n->value();
-
-    // check if we got scientific notation in result
-    if (ss.str().find_first_of("e") != std::string::npos) {
-      ss.clear(); ss.str(std::string());
-      ss.precision(std::max(12, opt.precision));
-      ss << std::fixed << n->value();
-    }
-
-    std::string tmp = ss.str();
-    size_t pos_point = tmp.find_first_of(".,");
-    size_t pos_fract = tmp.find_last_not_of("0");
-    bool is_int = pos_point == pos_fract ||
-                  pos_point == std::string::npos;
-
-    // reset stream for another run
-    ss.clear(); ss.str(std::string());
-
-    // take a shortcut for integers
-    if (is_int)
-    {
-      ss.precision(0);
-      ss << std::fixed << n->value();
-      res = std::string(ss.str());
-    }
-    // process floats
-    else
-    {
-      // do we have have too much precision?
-      if (pos_fract < opt.precision + pos_point)
-      { ss.precision((int)(pos_fract - pos_point)); }
-      else { ss.precision(opt.precision); }
-      // round value again
-      ss << std::fixed << n->value();
-      res = std::string(ss.str());
-      // maybe we truncated up to decimal point
-      size_t pos = res.find_last_not_of("0");
-      // handle case where we have a "0"
-      if (pos == std::string::npos) {
-        res = "0.0";
-      } else {
-        bool at_dec_point = res[pos] == '.' ||
-                            res[pos] == ',';
-        // don't leave a blank point
-        if (at_dec_point) ++ pos;
-        res.resize (pos + 1);
-      }
-    }
-
-    // some final cosmetics
-    if (res == "0.0") res = "0";
-    else if (res == "") res = "0";
-    else if (res == "-0") res = "0";
-    else if (res == "-0.0") res = "0";
-    else if (opt.output_style == COMPRESSED)
-    {
-      // check if handling negative nr
-      size_t off = res[0] == '-' ? 1 : 0;
-      // remove leading zero from floating point in compressed mode
-      if (n->zero() && res[off] == '0' && res[off+1] == '.') res.erase(off, 1);
-    }
-
-    // add unit now
-    res += n->unit();
-
+    // use values to_string facility
+    bool compressed = ctx->output_style == COMPRESSED;
+    std::string res = n->to_string(compressed, (int)ctx->precision);
     // output the final token
     append_token(res, n);
   }
 
-  // helper function for serializing colors
-  template <size_t range>
-  static double cap_channel(double c) {
-    if      (c > range) return range;
-    else if (c < 0)     return 0;
-    else                return c;
-  }
-
   void Inspect::operator()(Color* c)
   {
+    // use values to_string facility
+    bool compressed = ctx->output_style == COMPRESSED;
+	std::string res = c->to_string(compressed, (int)ctx->precision);
     // output the final token
-    std::stringstream ss;
-
-    // original color name
-    // maybe an unknown token
-    std::string name = c->disp();
-
-    // resolved color
-    std::string res_name = name;
-
-    double r = Sass::round(cap_channel<0xff>(c->r()), opt.precision);
-    double g = Sass::round(cap_channel<0xff>(c->g()), opt.precision);
-    double b = Sass::round(cap_channel<0xff>(c->b()), opt.precision);
-    double a = cap_channel<1>   (c->a());
-
-    // get color from given name (if one was given at all)
-    if (name != "" && name_to_color(name)) {
-      const Color* n = name_to_color(name);
-      r = Sass::round(cap_channel<0xff>(n->r()), opt.precision);
-      g = Sass::round(cap_channel<0xff>(n->g()), opt.precision);
-      b = Sass::round(cap_channel<0xff>(n->b()), opt.precision);
-      a = cap_channel<1>   (n->a());
-    }
-    // otherwise get the possible resolved color name
-    else {
-      double numval = r * 0x10000 + g * 0x100 + b;
-      if (color_to_name(numval))
-        res_name = color_to_name(numval);
-    }
-
-    std::stringstream hexlet;
-    bool compressed = opt.output_style == COMPRESSED;
-    hexlet << '#' << std::setw(1) << std::setfill('0');
-    // create a short color hexlet if there is any need for it
-    if (compressed && is_color_doublet(r, g, b) && a == 1) {
-      hexlet << std::hex << std::setw(1) << (static_cast<unsigned long>(r) >> 4);
-      hexlet << std::hex << std::setw(1) << (static_cast<unsigned long>(g) >> 4);
-      hexlet << std::hex << std::setw(1) << (static_cast<unsigned long>(b) >> 4);
-    } else {
-      hexlet << std::hex << std::setw(2) << static_cast<unsigned long>(r);
-      hexlet << std::hex << std::setw(2) << static_cast<unsigned long>(g);
-      hexlet << std::hex << std::setw(2) << static_cast<unsigned long>(b);
-    }
-
-    if (compressed && !c->is_delayed()) name = "";
-    if (opt.output_style == INSPECT && a >= 1) {
-      append_token(hexlet.str(), c);
-      return;
-    }
-
-    // retain the originally specified color definition if unchanged
-    if (name != "") {
-      ss << name;
-    }
-    else if (r == 0 && g == 0 && b == 0 && a == 0) {
-        ss << "transparent";
-    }
-    else if (a >= 1) {
-      if (res_name != "") {
-        if (compressed && hexlet.str().size() < res_name.size()) {
-          ss << hexlet.str();
-        } else {
-          ss << res_name;
-        }
-      }
-      else {
-        ss << hexlet.str();
-      }
-    }
-    else {
-      ss << "rgba(";
-      ss << static_cast<unsigned long>(r) << ",";
-      if (!compressed) ss << " ";
-      ss << static_cast<unsigned long>(g) << ",";
-      if (!compressed) ss << " ";
-      ss << static_cast<unsigned long>(b) << ",";
-      if (!compressed) ss << " ";
-      ss << a << ')';
-    }
-
-    append_token(ss.str(), c);
-
+    append_token(res, c);
   }
 
   void Inspect::operator()(Boolean* b)
   {
+    // use values to_string facility
+    bool compressed = ctx->output_style == COMPRESSED;
+	std::string res = b->to_string(compressed, (int)ctx->precision);
     // output the final token
-    append_token(b->value() ? "true" : "false", b);
+    append_token(res, b);
   }
 
   void Inspect::operator()(String_Schema* ss)
@@ -703,28 +496,25 @@ namespace Sass {
 
   void Inspect::operator()(String_Constant* s)
   {
-    append_token(s->value(), s);
+    // get options from optional? context
+    int precision = ctx ? (int)ctx->precision : 5;
+    bool compressed = ctx ? ctx->output_style == COMPRESSED : false;
+    // use values to_string facility
+    std::string res(s->to_string(compressed, precision));
+    // output the final token
+    append_token(res, s);
   }
 
   void Inspect::operator()(String_Quoted* s)
   {
-    if (const char q = s->quote_mark()) {
-      append_token(quote(s->value(), q), s);
-    } else {
-      append_token(s->value(), s);
-    }
+    // get options from optional? context
+    int precision = ctx ? (int)ctx->precision : 5;
+    bool compressed = ctx ? ctx->output_style == COMPRESSED : false;
+    // use values to_string facility
+    std::string res(s->to_string(compressed, precision));
+    // output the final token
+    append_token(res, s);
   }
-
-  void Inspect::operator()(Custom_Error* e)
-  {
-    append_token(e->message(), e);
-  }
-
-  void Inspect::operator()(Custom_Warning* w)
-  {
-    append_token(w->message(), w);
-  }
-
   void Inspect::operator()(Supports_Operator* so)
   {
 
@@ -821,8 +611,11 @@ namespace Sass {
 
   void Inspect::operator()(Null* n)
   {
+    // use values to_string facility
+    bool compressed = ctx->output_style == COMPRESSED;
+    std::string res = n->to_string(compressed, (int)ctx->precision);
     // output the final token
-    append_token("null", n);
+    append_token(res, n);
   }
 
   // parameters and arguments
@@ -945,10 +738,7 @@ namespace Sass {
     in_wrapped = true;
     append_token(s->name(), s);
     append_string("(");
-    bool was_comma_array = in_comma_array;
-    in_comma_array = false;
     s->selector()->perform(this);
-    in_comma_array = was_comma_array;
     append_string(")");
     in_wrapped = was;
   }
@@ -1025,52 +815,16 @@ namespace Sass {
 
   void Inspect::operator()(Selector_List* g)
   {
-
-    if (g->empty()) {
-      if (output_style() == TO_SASS) {
-        append_token("()", g);
-      }
-      return;
-    }
-
-
-    bool was_comma_array = in_comma_array;
-    // probably ruby sass eqivalent of element_needs_parens
-    if (output_style() == TO_SASS && g->length() == 1 &&
-      (!dynamic_cast<List*>((*g)[0]) &&
-       !dynamic_cast<Selector_List*>((*g)[0]))) {
-      append_string("(");
-    }
-    else if (!in_declaration && in_comma_array) {
-      append_string("(");
-    }
-
-    if (in_declaration) in_comma_array = true;
-
+    if (g->empty()) return;
     for (size_t i = 0, L = g->length(); i < L; ++i) {
       if (!in_wrapped && i == 0) append_indentation();
       if ((*g)[i] == 0) continue;
-      schedule_mapping((*g)[i]->last());
-      // add_open_mapping((*g)[i]->last());
       (*g)[i]->perform(this);
-      // add_close_mapping((*g)[i]->last());
       if (i < L - 1) {
         scheduled_space = 0;
         append_comma_separator();
       }
     }
-
-    in_comma_array = was_comma_array;
-    // probably ruby sass eqivalent of element_needs_parens
-    if (output_style() == TO_SASS && g->length() == 1 &&
-      (!dynamic_cast<List*>((*g)[0]) &&
-       !dynamic_cast<Selector_List*>((*g)[0]))) {
-      append_string(",)");
-    }
-    else if (!in_declaration && in_comma_array) {
-      append_string(")");
-    }
-
   }
 
   void Inspect::fallback_impl(AST_Node* n)
